// Convolution class for convolution of two RGB images on the GPU.
//
// Convolution: IFFT { FFT(RGB) * FFT(RGB) }
// Caution: This operation is expensive. It uses nine 2D FFTs in 
// five passes and a number of swizzling passes.
//
// Assumes that GLEW is available and has been initialized.
//
// Two RGB images in OpenGL textures are expected as input.
//
// Code written by Jeppe Revall Frisvad, 2009
// Copyright (c) DTU Informatics 2009

#include <iostream>
#include <GL/glew.h>
#include "FFT.h"
#include "load_shaders.h"
#include "Convolution.h"

using namespace std;

namespace
{
  GLchar* minimal_vert =
    "#version 120\n"
    "void main()\n"
    "{\n"
    "  gl_Position = ftransform();\n"
    "}";

  GLchar* r_frag =
    "#version 120\n"
    "#extension GL_ARB_texture_rectangle : enable            \n"
    "uniform sampler2DRect input1;                           \n"
    "uniform sampler2DRect input2;                           \n"
    "void main()                                             \n"
    "{                                                       \n"
    "  float in1 = texture2DRect(input1, gl_FragCoord.xy).r; \n"
    "  float in2 = texture2DRect(input2, gl_FragCoord.xy).r; \n"
    "  gl_FragColor.r = in1;                                 \n"
    "  gl_FragColor.b = in2;                                 \n" 
    "}";

  GLchar* g_frag =
    "#version 120\n"
    "#extension GL_ARB_texture_rectangle : enable            \n"
    "uniform sampler2DRect input1;                           \n"
    "uniform sampler2DRect input2;                           \n"
    "void main()                                             \n"
    "{                                                       \n"
    "  float in1 = texture2DRect(input1, gl_FragCoord.xy).g; \n"
    "  float in2 = texture2DRect(input2, gl_FragCoord.xy).g; \n"
    "  gl_FragColor.r = in1;                                 \n"
    "  gl_FragColor.b = in2;                                 \n" 
    "}";

  GLchar* b_frag =
    "#version 120\n"
    "#extension GL_ARB_texture_rectangle : enable            \n"
    "uniform sampler2DRect input1;                           \n"
    "uniform sampler2DRect input2;                           \n"
    "void main()                                             \n"
    "{                                                       \n"
    "  float in1 = texture2DRect(input1, gl_FragCoord.xy).b; \n"
    "  float in2 = texture2DRect(input2, gl_FragCoord.xy).b; \n"
    "  gl_FragColor.r = in1;                                 \n"
    "  gl_FragColor.b = in2;                                 \n" 
    "}";

  GLchar* rg_product_frag =
    "#version 120\n"
    "#extension GL_ARB_texture_rectangle : enable            \n"
    "uniform sampler2DRect input1;                            \n"
    "void main()                                             \n"
    "{                                                       \n"
    "  vec4 fft = texture2DRect(input1, gl_FragCoord.xy);     \n"
    "  gl_FragColor = vec4(fft.r*fft.b - fft.g*fft.a,        \n"
    "                      fft.r*fft.a + fft.g*fft.b,        \n"
    "                      0.0, 0.0);                        \n"
    "}";

  GLchar* ba_product_frag =
    "#version 120\n"
    "#extension GL_ARB_texture_rectangle : enable            \n"
    "uniform sampler2DRect input1;                            \n"
    "void main()                                             \n"
    "{                                                       \n"
    "  vec4 fft = texture2DRect(input1, gl_FragCoord.xy);     \n"
    "  gl_FragColor = vec4(0.0, 0.0,                         \n"
    "                      fft.r*fft.b - fft.g*fft.a,        \n"
    "                      fft.r*fft.a + fft.g*fft.b);       \n"
    "}";

  GLchar* output_frag =
    "#version 120\n"
    "#extension GL_ARB_texture_rectangle : enable            \n"
    "uniform sampler2DRect output1;                           \n"
    "void main()                                             \n"
    "{                                                       \n"
    "  gl_FragColor = texture2DRect(output1, gl_FragCoord.xy);\n"
    "}";
}

Convolution::Convolution(unsigned int width, unsigned int height,
                         unsigned int in_tex1, unsigned int in_tex2)
  : w(width), h(height), input1(in_tex1), input2(in_tex2), fft(0)
{
  if(!glIsTexture(in_tex1) || !glIsTexture(in_tex2))
  {
    cerr << "Error: Convolution initialized with invalid texture handles." << endl;
    exit(0);
  }
  init_shaders();
  init_texture(fft_in_tex);
  init_texture(fft_out1_tex);
  init_texture(fft_out2_tex);
  init_framebuffer();
  fft = new FFT(w, h, fft_in_tex);
}

void Convolution::init_texture(unsigned int& tex)
{
  glGenTextures(1, &tex);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, tex);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA32F_ARB, w, h, 0, GL_RGBA, GL_FLOAT, 0);        
}

void Convolution::init_framebuffer()
{
  glGenFramebuffersEXT(1, &fbo);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, fft_in_tex, 0);
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_RECTANGLE_ARB, fft_out1_tex, 0);
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_RECTANGLE_ARB, fft_out2_tex, 0);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);  
}

void Convolution::init_shaders()
{
  load_shaders(minimal_vert, r_frag, r_prog);
  load_shaders(minimal_vert, g_frag, g_prog);
  load_shaders(minimal_vert, b_frag, b_prog);
  load_shaders(minimal_vert, rg_product_frag, rg_product_prog);
  load_shaders(minimal_vert, ba_product_frag, ba_product_prog);
  load_shaders(minimal_vert, output_frag, out_prog);
  glUseProgram(r_prog);
  glUniform1i(glGetUniformLocation(r_prog, "input1"), 0);
  glUniform1i(glGetUniformLocation(r_prog, "input2"), 1);       
  glUseProgram(g_prog);
  glUniform1i(glGetUniformLocation(g_prog, "input1"), 0);
  glUniform1i(glGetUniformLocation(g_prog, "input2"), 1);       
  glUseProgram(b_prog);
  glUniform1i(glGetUniformLocation(b_prog, "input1"), 0);
  glUniform1i(glGetUniformLocation(b_prog, "input2"), 1);       
  glUseProgram(rg_product_prog);
  glUniform1i(glGetUniformLocation(rg_product_prog, "input1"), 0);
  glUseProgram(ba_product_prog);
  glUniform1i(glGetUniformLocation(ba_product_prog, "input1"), 0);
  glUseProgram(out_prog);
  glUniform1i(glGetUniformLocation(out_prog, "output1"), 0);
  glUseProgram(0);
}

void Convolution::set_projection() const
{
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0.0, static_cast<double>(w), 0.0, static_cast<double>(h), -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
}

void Convolution::unset_projection() const
{
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
}

void Convolution::draw_quad() const
{
  glBegin(GL_POLYGON);
    glVertex2f(0.0f, 0.0f);
    glVertex2f(static_cast<float>(w), 0.0f);
    glVertex2f(static_cast<float>(w), static_cast<float>(h));
    glVertex2f(0.0f, static_cast<float>(h));
  glEnd();
}

void Convolution::enable_texs() const
{
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, input1);
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glActiveTexture(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, input2);
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glActiveTexture(GL_TEXTURE0);  
}

void Convolution::disable_texs() const
{
  glActiveTexture(GL_TEXTURE1);
  glDisable(GL_TEXTURE_RECTANGLE_ARB);
  glActiveTexture(GL_TEXTURE0);
  glDisable(GL_TEXTURE_RECTANGLE_ARB);
}

// Convolution: IFFT { FFT(RGB) * FFT(RGB) }
// Caution: This function uses 9 2D FFTs in 5 passes and a number of swizzling passes
void Convolution::convolve()
{
  int vp[4];
  glGetIntegerv(GL_VIEWPORT, vp);
  glViewport(0, 0, w, h);
  set_projection();

  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
  glClear(GL_COLOR_BUFFER_BIT);
  glBlendFunc(GL_ONE, GL_ONE);

  // Pass red bands as input for FFT
  glUseProgram(r_prog);
  enable_texs();
  draw_quad();
  disable_texs();
  fft->redraw_input();
  fft->do_fft();

  // Render complex multiplication of FFT result to output texture
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glUseProgram(rg_product_prog);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, fft->get_output());
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);
  draw_quad();
  glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

  // Pass green bands as input for FFT
  glUseProgram(g_prog);
  enable_texs();
  draw_quad();
  disable_texs();
  fft->redraw_input();
  fft->do_fft();

  // Render complex multiplication of FFT result to output texture
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glUseProgram(ba_product_prog);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, fft->get_output());
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);
  glEnable(GL_BLEND);
  draw_quad();
  glDisable(GL_BLEND);
  glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

  // Use the multiplied FFTs to convolve the input images
  glUseProgram(out_prog);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, fft_out1_tex);
  draw_quad();
  fft->invert(1);
  fft->invert(2);
  fft->redraw_input();
  fft->do_fft();
  fft->invert(1);
  fft->invert(2);

  //// Render inverse FFT result to output texture
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT);
  fft->draw_output(1.0f, 0.0f, 0.0f, 1);
  glEnable(GL_BLEND);
  fft->draw_output(0.0f, 1.0f, 0.0f, 2);
  glDisable(GL_BLEND);
  glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

  // Pass blue bands as input for FFT
  glUseProgram(b_prog);
  enable_texs();
  draw_quad();
  disable_texs();
  fft->redraw_input();
  fft->do_fft();

  // Render complex multiplication of FFT result to output texture
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glUseProgram(rg_product_prog);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, fft->get_output());
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);
  draw_quad();
  glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

  // Use the multiplied FFT to convolve the input images
  glUseProgram(out_prog);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, fft_out1_tex);  
  draw_quad();
  fft->invert(1);
  fft->invert(2);
  fft->redraw_input();
  fft->do_fft();
  fft->invert(1);
  fft->invert(2);

  // Render inverse FFT result to output texture
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT);
  glEnable(GL_BLEND);
  fft->draw_output(0.0f, 0.0f, 1.0f, 1);
  glDisable(GL_BLEND);
  glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

  // Reset states
  glUseProgram(0);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
  unset_projection();
  glViewport(vp[0], vp[1], vp[2], vp[3]);
}

void Convolution::draw_output() const
{
  set_projection();
  glUseProgram(out_prog);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, fft_out2_tex);
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  draw_quad();
  glDisable(GL_TEXTURE_RECTANGLE_ARB);
  glUseProgram(0);
  unset_projection();
}
