// Computation of glare from a candle light as described
// in the publication:
//
// Ritschel, T., Ihrke, M., Frisvad, J. R., Coppens, J., 
// Myszkowski, K., and Seidel, H.-P. Temporal glare:
// Real-time dynamic simulation of the scattering in the
// human eye. Computer Graphics Forum (Proceedings of 
// Eurographics 2009) 28(2), 183-192, 2009.
//
// This demo uses the following software packages:
// GLUT, GLEW, SOIL, gpu_convolution, spectrum_to_3stimuli,
// simplex, load_shaders, luminance, Timer.
// 
// Code written by Jeppe Revall Frisvad, August 2009.
// Copyright (c) DTU Informatics 2009

#include <iostream>
#include <sstream>
#include <cmath>
#include <GL/glew.h>
#include <GL/glut.h>
#include "../SOIL/SOIL.h"
#include "load_shaders.h"
#include "spectrum2rgb.h"
#include "simplex.h"
#include "FFT.h"
#include "Convolution.h"
#include "luminance.h"
#include "Timer.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

using namespace std;

//////////////////////////////////////////////////////////////////////
// Globals
//////////////////////////////////////////////////////////////////////

namespace
{
  int winw = 512, winh = 512;
  const unsigned int FFT_SIZE = 512;
  const unsigned int cornea_particles = 200;
  const unsigned int lens_particles = 400;
  const unsigned int lens_fibers = 200;
  const double lens_nucleus = 0.4;
  const float glare_intensity = 1.0f/static_cast<float>(FFT_SIZE*spectrum_rgb_samples);
  const float image_intensity = 10.0;
  const float particle_scale = 0.5;
  const float glare_scale = 1.5f;
  const int sim_timer = 20;
  float pupil_size = 7.0f;
  float contrast = 0.0f;
  float lambda1;
  double field_luminance = 1.0;

  FFT* fft;
  Convolution* convolution;
  bool draw_fft = true; 
  bool redrawn = false;
  bool tone_map = true;
  bool convolve = true;
  GLhandleARB pupil_prog;
  GLhandleARB tone_prog;
  unsigned int cornea_list;
  unsigned int lens_list;
  unsigned int source_list;
  unsigned int image_list;
  unsigned int source_tex;
  unsigned int image_tex;
  unsigned int final_tex;
  unsigned int fbo;
  void (*draw_func)() = 0;

  FrameRateTimer timer;
  bool measure_fps = false;

  double my_random(double from, double to)
  {
    return (to - from)*rand()/static_cast<double>(RAND_MAX) + from;
  }

  void simulate_hippus(double time_elapsed)
  {
    static const int no_of_octaves = 3;
    static const double pupil_max = 9.0;
    static const double amp = pupil_max/pupil_size*sqrt(1.0 - pupil_size/pupil_max)*contrast;
    static const double scale = 1.0/pupil_size;   // scale is the number of blobs in an interval of width 2
    static double x = 0.0;

    x += time_elapsed;
    double y = pupil_size;
    double octave = 1.0;
    double noise_x = x*scale;
    for(int j = 0; j < no_of_octaves; ++j)
    {
      double nx = noise_x*octave;
      y += amp/octave*Classic::noise(nx + 3.5, nx + 2.1, nx + 7.2);
      octave *= 2.0;
    }
    glUseProgram(pupil_prog);
    float pupil_radius = static_cast<float>(y)/10.0f;       // normalized pupil radius (image width: N = 10 mm)
    glUniform1f(glGetUniformLocation(pupil_prog, "pupil"), pupil_radius*pupil_radius);
    glUseProgram(0);
  }

//////////////////////////////////////////////////////////////////////
// Vertex shaders
//////////////////////////////////////////////////////////////////////

  GLcharARB* minimal_vert =
    "#version 120\n"
    "void main()\n"
    "{\n"
    "  gl_Position = ftransform();\n"
    "}";

  GLcharARB* position_vert =
    "#version 120\n"
    "varying vec4 color;\n"
    "varying vec2 position;\n"
    "void main()\n"
    "{\n"
    "  color = gl_Color;\n"
    "  position = (gl_ModelViewMatrix*gl_Vertex).xy;\n"
    "  gl_Position = ftransform();\n"
    "}";

//////////////////////////////////////////////////////////////////////
// Fragment shaders
//////////////////////////////////////////////////////////////////////

  GLchar* minimal_frag =
    "#version 120\n"
    "void main()\n"
    "{\n"
    "  gl_FragColor = vec4(1.0, 0.0, 0.0, 0.0);\n"
    "}";

  GLchar* circle_frag =
    "#version 120\n"
    "varying vec4 color;\n"
    "varying vec2 position;\n"
    "uniform float coef;\n"   // coefficient for E
    "uniform float pupil;\n"
    "void main()\n"
    "{\n"
    "  float sqr_dist = dot(position, position);\n"
    "  if(sqr_dist > pupil) discard;\n"

    "  vec2 E = vec2(cos(coef*sqr_dist), sin(coef*sqr_dist));\n"  
    //"  sincos(coef*sqr_dist, E.y, E.x);\n" // Find E (alternatively use texture for speed-up)
    "  vec2 P = color.rg;\n"       
    "  gl_FragColor.rg = vec2(P.x*E.x - P.y*E.y, P.x*E.y + P.y*E.x);\n"  
    "  P = color.ba;\n"       
    "  gl_FragColor.ba = vec2(P.x*E.x - P.y*E.y, P.x*E.y + P.y*E.x);\n"  
    "}";

  GLchar* tone_frag =
    "#version 120\n"
    "#extension GL_ARB_texture_rectangle : enable\n"
    "uniform sampler2DRect final;\n"
    "uniform float max_intensity;\n"
    "void main()\n"
    "{\n"
    "  vec4 color = texture2DRect(final, gl_FragCoord.xy);\n"
    "  color.rgb = color.rgb/max_intensity;\n"
    "  color = pow(color, vec4(0.4545));\n"
    "  gl_FragColor = color;\n"
    "}";
}

void simulate(int value);

//////////////////////////////////////////////////////////////////////
// Draw functions
//////////////////////////////////////////////////////////////////////

void set_projection()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
}

void draw_pupil()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glUseProgram(pupil_prog);
  glCallList(cornea_list);
  glCallList(lens_list);
  glUseProgram(0);
}

void draw_source()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glCallList(source_list);
}

void draw_image()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glCallList(image_list); 
}

//////////////////////////////////////////////////////////////////////
// Initialization
//////////////////////////////////////////////////////////////////////

void GLEWinit()
{
  GLenum err = glewInit();
  if(err != GLEW_OK)
  {
    cerr << "Error loading GLEW: " << glewGetErrorString(err) << endl;
    exit(0);
  }

  if(!GLEW_VERSION_2_1)
  {
    cerr << "Warning: OpenGL version 2.1 is not supported" << endl;
  }

  if(!GLEW_ARB_texture_rectangle)
  {
    cerr << "Error: Rectangular textures are not supported" << endl;
    exit(0);
  }

  if(!GLEW_ARB_fragment_program)
  {
    cerr << "Error: fragment programs are not supported" << endl;
    exit(0);
  }

  if(!GLEW_ARB_texture_float)
  {
    cerr << "Error: floating point textures are not supported" << endl;
    exit(0);
  }

  if(!GLEW_EXT_framebuffer_object)
  {
    cerr << "Error: framebuffer objects are not supported" << endl;
    exit(0);
  }
}

void GLinit()
{
  glEnable(GL_CULL_FACE);
  glEnable(GL_POINT_SMOOTH);
}

void init_cornea()
{
  srand(41);
  cornea_list = glGenLists(1);
  glNewList(cornea_list, GL_COMPILE);

  // Draw pupil (aperture)
  glColor4f(1.0f, 0.0f, 0.0f, 0.0f);
  glBegin(GL_POLYGON);
    glVertex2f(-1.0f, -1.0f);
    glVertex2f( 1.0f, -1.0f);
    glVertex2f( 1.0f,  1.0f);
    glVertex2f(-1.0f,  1.0f);
  glEnd();

  // Draw particles (occluders)
  glColor4f(0.0f, 0.0f, 0.0f, 0.0f);
  glPointSize(12.0f*particle_scale);
  glBegin(GL_POINTS);
  for(unsigned int i = 0; i < cornea_particles; ++i)
    glVertex2d(my_random(-1.0, 1.0), my_random(-1.0, 1.0));
  glEnd();

  glEndList();
}

void init_lens()
{
  srand(43);
  lens_list = glGenLists(1);
  glNewList(lens_list, GL_COMPILE);

  // Draw lens nucleus particles (occluders)
  glColor4f(0.0f, 0.0f, 0.0f, 0.0f);
  glPointSize(4.0f*particle_scale);
  glBegin(GL_POINTS);
  for(unsigned int i = 0; i < lens_particles; ++i)
  {
    double x, y;
    do
    {
      x = my_random(-lens_nucleus, lens_nucleus);
      y = my_random(-lens_nucleus, lens_nucleus);
    }
    while(sqrt(x*x + y*y) > lens_nucleus);
    glVertex2d(x, y);
  }
  glEnd();

  // Draw lens cortex fiber membranes (grating)
  double fiber_spacing = 2.0*M_PI/static_cast<double>(lens_fibers);

  glLineWidth(3.0f*particle_scale);
  glBegin(GL_LINES);
  for(unsigned int i = 0; i < lens_fibers; ++i)
  {
    double angle = i*fiber_spacing;
    double x = cos(angle);
    double y = sin(angle);
    glColor4f(0.0f, 0.0f, 0.0f, 0.0f);
    glVertex2d(x, y);
    glColor4f(1.0f, 0.0f, 0.0f, 0.0f);
    glVertex2d(x*lens_nucleus, y*lens_nucleus);
  }
  glEnd();

  glEndList();
}

void init_source(const char* imagefile, const char* overlayfile)
{
  float w = static_cast<float>(winw);
  float h = static_cast<float>(winh);
  
  // Load image file for background
  image_tex = SOIL_load_OGL_texture(imagefile, SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_INVERT_Y | SOIL_FLAG_TEXTURE_RECTANGLE);

  image_list = glGenLists(1);
  glNewList(image_list, GL_COMPILE);

  glBlendFunc(GL_ONE, GL_ONE);
  glEnable(GL_BLEND);
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, image_tex);
  glColor3f(1.0f, 1.0f, 1.0f);
  glBegin(GL_POLYGON);
    glTexCoord2f(0.0f,  0.0f);
    glVertex2f( -1.0f, -1.0f);
    glTexCoord2f(   w,  0.0f);
    glVertex2f(  1.0f, -1.0f);
    glTexCoord2f(   w,     h);
    glVertex2f(  1.0f,  1.0f);
    glTexCoord2f(0.0f,     h);
    glVertex2f( -1.0f,  1.0f);
  glEnd();
  glDisable(GL_TEXTURE_RECTANGLE_ARB);
  glDisable(GL_BLEND);

  glEndList();

  // Load overlay file which becomes the HDR light source
  int width, height, channels;
  unsigned char* data = SOIL_load_image(overlayfile, &width, &height, &channels, SOIL_LOAD_RGB);
  if(!data)
  {
    cerr << "Error: could not load source image" << endl;
    exit(0);
  }

  // Compute field luminance
  field_luminance = 0.0;
  double max_luminance = 0.0;
  for(int i = 0; i < height; ++i)
    for(int j = 0; j < width; ++j)
    {
      int idx = (i*width + j)*3;
      double r = (data[idx] + 0.5)/256.0;
      double g = (data[idx + 1] + 0.5)/256.0;
      double b = (data[idx + 2] + 0.5)/256.0;
      float luminance = static_cast<float>(get_luminance_NTSC(r, g, b));
      field_luminance += luminance;
      max_luminance = luminance > max_luminance ? luminance : max_luminance;
    }
  field_luminance /= static_cast<double>(width*height);
  contrast = static_cast<float>((max_luminance - field_luminance)/max_luminance); // use a factor here to dramatise the effect
  
  // Compute pupil size based on field luminance
  pupil_size = static_cast<float>(4.9 - 3.0*tanh(0.4*(log10(field_luminance) + 1.0)));

  source_tex = SOIL_create_OGL_texture(
    data, width, height, channels,
    SOIL_CREATE_NEW_ID, 
    SOIL_FLAG_POWER_OF_TWO | SOIL_FLAG_INVERT_Y | SOIL_FLAG_TEXTURE_RECTANGLE);
  if(!source_tex)
  {
    cerr << "Error: could not create texture from source image" << endl;
    exit(0);
  }

  source_list = glGenLists(1);
  glNewList(source_list, GL_COMPILE);

  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, source_tex);
  glColor3f(1.0f, 1.0f, 1.0f);
  glBegin(GL_POLYGON);
    glTexCoord2f(0.0f,  0.0f);
    glVertex2f( -1.0f, -1.0f);
    glTexCoord2f(   w,  0.0f);
    glVertex2f(  1.0f, -1.0f);
    glTexCoord2f(   w,     h);
    glVertex2f(  1.0f,  1.0f);
    glTexCoord2f(0.0f,     h);
    glVertex2f( -1.0f,  1.0f);
  glEnd();
  glDisable(GL_TEXTURE_RECTANGLE_ARB);

  glEndList();
}

void init_framebuffer()
{
  glGenTextures(1, &final_tex);
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, final_tex);
  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB32F_ARB, winw, winh, 0, GL_RGB, GL_FLOAT, 0);        

  glGenFramebuffersEXT(1, &fbo);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, final_tex, 0);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);  
}

void init(const char* imagefile, const char* overlayfile)
{
  GLEWinit();
  GLinit();
  init_cornea();
  init_lens();
  init_source(imagefile, overlayfile);
  init_framebuffer();

  load_shaders(position_vert, circle_frag, pupil_prog);
  load_shaders(minimal_vert, tone_frag, tone_prog);
  glUseProgram(pupil_prog);
  float pupil_radius = pupil_size/10.0f;       // normalized pupil radius (image width: N = 10 mm)
  float a = 10.0f/4.0f;                        // denormalizing pupil plane distances (N/2*N/2)/N = N/4
  glUniform1f(glGetUniformLocation(pupil_prog, "coef"), a*static_cast<float>(M_PI));
  glUniform1f(glGetUniformLocation(pupil_prog, "pupil"), pupil_radius*pupil_radius);
  glUseProgram(0);

  float d = 20.32f;                            // distance from pupil to retina [mm]
  lambda1 = 10.0f/d*1.0e3f;                    // wavelength in micrometers

  fft = new FFT(FFT_SIZE, FFT_SIZE);
  fft->set_input(draw_pupil);
  fft->do_fft();

  convolution = new Convolution(FFT_SIZE, FFT_SIZE, source_tex, final_tex);

  draw_func = draw_pupil;
}

//////////////////////////////////////////////////////////////////////
// GLUT Callback Functions
//////////////////////////////////////////////////////////////////////

void display()
{
  static bool first = true;

  if(first)
  {
    first = false;
    glutTimerFunc(sim_timer, simulate, 0);
  }

  glClear(GL_COLOR_BUFFER_BIT);
  glLoadIdentity();

  if(draw_fft)
  {
    if(fft->newly_redrawn())
      fft->draw_output(1.0f, 1.0f, 1.0f); 
    else
    {
      if(tone_map || convolve)
      {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
        glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
        glClear(GL_COLOR_BUFFER_BIT);
      }      
      glBlendFunc(GL_ONE, GL_ONE);
      glEnable(GL_BLEND);
      float move = static_cast<float>(FFT_SIZE)*0.5f;
      for(int i = 0; i < spectrum_rgb_samples; ++i)
      {
        float scale = static_cast<float>(spectrum_rgb[i][0])/lambda1; // image scale = 1/(coordinate scale)
        glLoadIdentity();
        glTranslatef(move, move, 0.0f);
        glScalef(glare_scale*scale, glare_scale*scale, 1.0f);
        glTranslatef(-move, -move, 0.0f);
        fft->draw_output(static_cast<float>(spectrum_rgb[i][1])*glare_intensity/scale*1.0f, 
                         static_cast<float>(spectrum_rgb[i][2])*glare_intensity/scale*2.0f,
                         static_cast<float>(spectrum_rgb[i][3])*glare_intensity/scale*2.0f);
      }
      glDisable(GL_BLEND);
      glLoadIdentity();

      float max_intensity = 1.0f;
      if(convolve)
      {
        convolution->convolve();
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
        glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
        convolution->draw_output();        
        max_intensity *= FFT_SIZE;
        max_intensity /= image_intensity;

        // Compute intensity scale according to pupil size
        max_intensity *= glare_scale*glare_scale;
        float radius_sqr;
        glGetUniformfv(pupil_prog, glGetUniformLocation(pupil_prog, "pupil"), &radius_sqr);
        float pupil_area = radius_sqr*0.25f*static_cast<float>(M_PI); // unit circle to circle in 1x1 box
        max_intensity /= pupil_area;        
      }
      if(tone_map || convolve)
      {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);  

        set_projection();
        glLoadIdentity();
        glUseProgram(tone_prog);
        glUniform1f(glGetUniformLocation(tone_prog, "max_intensity"), max_intensity); 
        glEnable(GL_TEXTURE_RECTANGLE_ARB);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, final_tex);
        glBegin(GL_POLYGON);
          glVertex2f(-1.0f, -1.0f);
          glVertex2f( 1.0f, -1.0f);
          glVertex2f( 1.0f,  1.0f);
          glVertex2f(-1.0f,  1.0f);
        glEnd();
        glDisable(GL_TEXTURE_RECTANGLE_ARB);
        glUseProgram(0);
        if(convolve)
          glCallList(image_list);
      }
    }
  }
  else if(draw_func)
    draw_func();

  glutSwapBuffers();

  if(measure_fps)
  {
    ostringstream ostr;
    ostr << "Glare demo (fps: " << timer.get_fps() << ")";
    glutSetWindowTitle(ostr.str().c_str());
  }
}

void reshape(int width, int height)
{
  winw = width;
  winh = height;

  glViewport(0, 0, winw, winh);
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key)
  {
  // Press <space> to toggle between input image and glare result.
  case ' ':
    if(!convolve || !draw_fft || draw_func != draw_pupil)
    {
      draw_func = draw_pupil;
      fft->set_input(draw_func);
      fft->do_fft();
      draw_fft = true;
      convolve = true;
      break;
    }
  // Press '0' to show the input image.
  case '0':
    convolve = false;
    draw_fft = false;
    draw_func = draw_image;
    fft->set_input(draw_func);
    fft->do_fft();
    glutTimerFunc(sim_timer, simulate, 0);
    break;
  // Press '1' to show the source overlay (intense pixels causing glare).
  case '1':
    convolve = false;
    draw_fft = false;
    draw_func = draw_source;
    fft->set_input(draw_func);
    fft->do_fft();
    glutTimerFunc(sim_timer, simulate, 0);
    break;
  // Press '2' to show the pupil image to be used for computing glare.
  case '2':
    convolve = false;
    draw_fft = false;
    draw_func = draw_pupil;
    fft->set_input(draw_func);
    fft->do_fft();
    glutTimerFunc(sim_timer, simulate, 0);
    break;
  // Press 'c' to toggle on/off convolution of the source overlay with the point spread function.
  case 'c':
    convolve = !convolve;
    break;
  // Press 'f' to toggle on/off computation of the point spread function from the pupil image.
  case 'f':
    draw_fft = !draw_fft;
    break;
  // Press 'g' to toggle gamma correction on/off.
  case 'g':
    tone_map = !tone_map;
    break;
  // Press 'm' to measure the rendering frame rate.
  case 'm':
    measure_fps = !measure_fps;
    if(!measure_fps)
      glutSetWindowTitle("Glare demo");
    break;
  // Press <esc> to quit the program.
  case 27:
    delete fft;
    delete convolution;
    exit(0);
  }
  glutPostRedisplay();
}

void simulate(int value)
{
  if(fft->newly_redrawn())
    return;

  simulate_hippus(sim_timer*1.0e-3);
  fft->do_fft();
  glutTimerFunc(sim_timer, simulate, 0);  
  glutPostRedisplay();  
}

void idle()
{
  if(measure_fps)
    glutPostRedisplay();
}

//////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
  glutInit(&argc, argv);
  glutInitWindowSize(winw, winh);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA);
  glutCreateWindow("Glare demo");
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);
  glutIdleFunc(idle);

  init(argc > 1 ? argv[1] : "candle.png", argc > 2 ? argv[2] : "overlay.png");

  glutMainLoop();

  return 0;
}